/*
 * function [x,iterations,total_time,exec_time] = SpaRSA_gpu(y,x_start,dct_rows,tau,continuationSteps,stop_tol)
 *
 * GPU (CUDA) implementation of Compressed sensing algorithm, SpaRSA, for 1-D DCT sensing matrix.
 *
 * Author: Sangkyun Lee
 * Last Modified: 2008/11/7
 */

#include "mex.h"

#define BLOCK_SIZE 32
// comment out VERBOSE and SAFE_MODE for speed measurements
// #define VERBOSE
// #define SAFE_MODE	// CUDA_CALL checks error
#define MATLAB		// debug() works as mexPrint()

//#include "cudaMath.cu"	// included by the dct_gpu.cu in the next line
#include "../dct_gpu/dct_gpu.cu"

#define DISP_GPUVAR(x_g, type, n) { \
    type *x;	\
    x = (type*)malloc(sizeof(type)*n);\
    cudaMemcpy(x, x_g, sizeof(type)*n, cudaMemcpyDeviceToHost);\
    for(int i=0; i<20; ++i)\
	mexPrintf("%f \n", x[i]);\
    mexPrintf("\n");\
    free(x);\
}

#define REALMIN 1.1755e-38f

void print_usage()
{
    mexPrintf("function [x, iterations, total_time, exec_time] = SpaRSA_gpu(y,x_start,dct_rows,tau,continuationSteps,stop_tol)\n"); 
}

/*
 * Compute A*x
 */
inline void A(FLOAT* out, FLOAT* in, int n, int k, int* rows)
{
    dct_gpu(out, in, n, k, rows);
}

/*
 * Compute A'*x
 */
inline void AT(FLOAT* out, FLOAT* in, int n, int k, int* rows)
{
    idct_gpu(out, in, n, k, rows);
}


/*
 * A function implements the following Matlab function
 *

function y = soft(x,T)
if sum(abs(T(:)))==0
   y = x;
else
   y = max(abs(x) - T, 0);
   y = y./(y+T) .* x;
end

 *
 */

__global__ void __soft(FLOAT* y, const FLOAT* x, FLOAT T, int m);
void soft(FLOAT* y, const FLOAT* x, FLOAT T, int m)
{
    dim3 dimBlock(BLOCK_SIZE, 1);
    dim3 dimGrid((int)fmaxf(1.f, ceilf(1.f*m/BLOCK_SIZE)), 1);

    __soft<<<dimGrid, dimBlock>>> (y, x, T, m);
}

__global__ void __soft(FLOAT* y, const FLOAT* x, FLOAT T, int m)
{
    int X = X_INDEX;
    FLOAT x_e, y_e;

    if(X < m)
    {
	x_e = x[X];
	y_e = fmaxf(fabsf(x_e) - T, 0.f);
	y[X] = y_e / (y_e + T) * x_e;
    }
}


/*
 * Main entry point for Matlab call
 */
//function [x, iterations, total_time, exec_time] = SpaRSA_gpu(y,x_start,dct_rows,tau,continuationSteps,stop_tol)
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[])
{
    double turn_time, exec_time;

    if(nrhs < 6) {
	print_usage();
	return;
    }

#ifdef VERBOSE
    debug("Warning: VERBOSE flag is on! Turn off this flag for speed measurements\n");
#endif
#ifdef SAFE_MODE
    debug("Warning: SAFEMODE flag is on! Turn off this flag for speed measurements\n");
#endif

    /*
     * Choose a GPU device to use.
     *
     * You can do this by using gpu_warmup() function instead.
     */
    // CUDA_CALL( cudaSetDevice(2) );

turn_time = cputime();

cudaStream_t st[2];
cudaStreamCreate(&st[0]);
cudaStreamCreate(&st[1]);
float* kkk;
CUDA_CALL( cudaMalloc((void**)&kkk, 4) );
CUDA_CALL( cudaFree(tmp) );

    /*
     * Retrieve matlab data
     */
    // Input
    FLOAT *x_h, *y_h, *x_start, tau, stop_tol;
    INT n, k, *dct_rows_h, continuationSteps;

    FLOAT *taus;	// tau values used in continuation
    INT nContinuation; // modified version of continuationSteps
    FLOAT Aty_max;	      		  	// scalar

    y_h = (FLOAT*) mxGetData(prhs[0]);			// of length k
    x_start = (FLOAT*) mxGetData(prhs[1]);		// of length n
    dct_rows_h = (INT*) mxGetData(prhs[2]);		// of length k
    tau = (FLOAT) mxGetScalar(prhs[3]);
    continuationSteps = (INT) mxGetScalar(prhs[4]);
    stop_tol = (FLOAT) mxGetScalar(prhs[5]);

    k = mxGetM(prhs[0]);
    n = mxGetM(prhs[1]);

//mexPrintf("n %d k %d\n", n, k);
//for(int i=0; i<k; ++i)
//    mexPrintf("%d ", dct_rows_h[i]);
//mexPrintf("\n");

    // define the data structure to contain tau values used in continuation
    if (continuationSteps<1) 
      nContinuation=1;
    else
      nContinuation=continuationSteps;
    taus = (FLOAT *)malloc(sizeof(float)*nContinuation);

    /*
     * GPU launch configuration
     */
    dim3 dimBlock(BLOCK_SIZE, 1);
    dim3 dimGrid_k((int)fmaxf(ceilf(1.f*k/BLOCK_SIZE),1.f), 1);
    dim3 dimGrid_n((int)fmaxf(ceilf(1.f*n/BLOCK_SIZE),1.f), 1);

    dct_gpu_init(n,1);
    cudaMath_init();

    /*
     * GPU variables
     */
    FLOAT *x, *prev_x, *dx, *Aty, *gradq;	// of length n
    FLOAT *y, *resid, *prev_resid, *Adx;	// of length k
    INT *dct_rows;				// of length k

    CUDA_CALL( cudaMalloc((void**)&x, sizeof(FLOAT)*n) );
    CUDA_CALL( cudaMalloc((void**)&prev_x, sizeof(FLOAT)*n) );
    CUDA_CALL( cudaMalloc((void**)&dx, sizeof(FLOAT)*n) );
    CUDA_CALL( cudaMalloc((void**)&Aty, sizeof(FLOAT)*n) );
    CUDA_CALL( cudaMalloc((void**)&gradq, sizeof(FLOAT)*n) );

    CUDA_CALL( cudaMalloc((void**)&y, sizeof(FLOAT)*k) );
    CUDA_CALL( cudaMalloc((void**)&resid, sizeof(FLOAT)*k) );
    CUDA_CALL( cudaMalloc((void**)&prev_resid, sizeof(FLOAT)*k) );
//    CUDA_CALL( cudaMalloc((void**)&w, sizeof(FLOAT)*k) ); // for duality test
    CUDA_CALL( cudaMalloc((void**)&Adx, sizeof(FLOAT)*k) );

    CUDA_CALL( cudaMalloc((void**)&dct_rows, sizeof(INT)*k) );

    CUDA_CALL( cudaMemcpy(x, x_start, sizeof(FLOAT)*n, cudaMemcpyHostToDevice) );
    CUDA_CALL( cudaMemcpy(y, y_h, sizeof(FLOAT)*k, cudaMemcpyHostToDevice) );
    CUDA_CALL( cudaMemcpy(dct_rows, dct_rows_h, sizeof(INT)*k, cudaMemcpyHostToDevice) );

    /*
     * Algorithm control parameters
     */
    INT maxiter = 1000;
    INT bbCycle = 1;		// cycle length for Barzilai-Borwein
    int enforceMonotone = 0;	// enforce monotonicity?
    int stopCriterion = 1;	// 1=function value test, 2=duality test
    FLOAT alphamin = 1e-30;
    FLOAT alphamax = 1e30;
    FLOAT eta = 2.0;		
       // increase alpha by this factor after unsuccessful step
//    FLOAT alphaFactor = 0.8;	
       // decrease factor between iterations, if BB not used.

    /*
     * Algorithm control variables
     */
    INT iter = 0;
    INT iterThisCycle = 0;
    FLOAT alpha = 1.0f;
    FLOAT f, prev_f, residTresid, prev_residTresid, norm1x, prev_norm1x;
    int keep_going, cont_inner;
    
///% Precompute A'*y since it'll be used a lot
///Aty = AT(y);
    AT(Aty, y, n, k, dct_rows);
/// ||Ayt||_inf useful in determining initial tau for continuation
    Aty_max = norm_inf(Aty, n);
// compute the roster of tau values for continuation
    if (tau >= 0.8*Aty_max || nContinuation==1) {
      taus[0] = tau;
    } else {
      // set the parameter sigma that controls how much the smaller values of
      // tau are bunched together. Need sigma>=1, larger values give tighter 
      // bunching.
      FLOAT sigma=1.0f;
      taus[0] = 0.8*Aty_max; taus[nContinuation-1]=tau;
      FLOAT logtau_0, logtau_m, logtau, tauFac;
      logtau_0 = log(taus[0]); logtau_m = log(taus[nContinuation-1]);
      for (int i=1; i<nContinuation-1; i++) {
        tauFac = (1.f*nContinuation-1.f-1.f*i) / (1.f*nContinuation-1.f);
	tauFac = pow(tauFac,sigma);
        logtau = logtau_m + tauFac*(logtau_0 - logtau_m);
        taus[i] = exp(logtau);
      }      
    }

exec_time = cputime();
iter=0;
// the outer (continuation) loop
  for (int j=0; j<nContinuation && iter<=maxiter; j++) {
    tau = taus[j];
    iterThisCycle=0;
    iter++;
    ///% Compute the initial residual and gradient
    ///resid =  A(x) - y;
    ///size(resid)
    ///gradq = AT(resid);
    A(resid, x, n, k, dct_rows);
    __minusMatMat<<<dimGrid_k, dimBlock>>> (resid, resid, y, k, 1);
    //cublasSaxpy(k, -1.f, y, 1, resid, 1);
    AT(gradq, resid, n, k, dct_rows);

    ///f = 0.5*(resid'*resid) + tau * norm(x,1);
    residTresid = inner(resid, resid, k);

    norm1x = norm_1(x,n);
    f = 0.5f*residTresid + tau*norm1x;
    //mexPrintf("f = %.8f\n", f);
#ifdef VERBOSE
     mexPrintf("\n continuation step %d, tau=%7.3e, iter=%4d, obj=%10.6e\n", 
        j, tau, iter, f);
#endif

    ///% control variable
    keep_going = 1;
    while(keep_going)
    {
      iter++;
      //gradq = AT(resid);
      AT(gradq, resid, n, k, dct_rows);

      // save current values
      // prev_x = x;
      // prev_f = f;
      // prev_resid = resid;
      CUDA_CALL( cudaMemcpy(prev_x, x, sizeof(FLOAT)*n, cudaMemcpyDeviceToDevice) );
      CUDA_CALL( cudaMemcpy(prev_resid, resid, sizeof(FLOAT)*k, cudaMemcpyDeviceToDevice) );
      prev_f = f;
      prev_residTresid = residTresid; prev_norm1x = norm1x;

      /* computation of step */
      cont_inner = 1;
      while(cont_inner)
      {
        //x = soft(prev_x - gradq*(1/alpha),tau/alpha);
 	__minusMatMat<<<dimGrid_n, dimBlock>>> (x, prev_x, 1.f, gradq, (1.f/alpha), n, 1);
	 soft(x, x, tau/alpha, n);

	 //resid = A(x)-y;
	 A(resid, x, n, k, dct_rows);
	 __minusMatMat<<<dimGrid_k, dimBlock>>> (resid, resid, y, k, 1);
	 //cublasSaxpy(k, -1.f, y, 1, resid, 1);

	 //f = 0.5*(resid(:)'*resid(:)) + tau * norm(x,1);
	 residTresid = inner(resid, resid, k);
	 norm1x = norm_1(x,n);
	 f = 0.5f*residTresid + tau*norm1x;

	 /* did the function decrease? do we care? */
	 if (f <= prev_f || !enforceMonotone)
	 {
	   cont_inner=0;
	 } else {
	   // not good enough, increase alpha and try again
	   alpha = eta*alpha;
	 }
       } // end of cont_inner
 
#ifdef VERBOSE
    mexPrintf("iter=%4d, obj=%10.6e, alpha=%6.3e ", iter, f, alpha );
#endif

    /* standard BB choice of initial alpha for next step */
    if (iterThisCycle==0 || enforceMonotone==1)
    {
	//dx = x - prev_x;
	__minusMatMat<<<dimGrid_n, dimBlock, 0, st[0]>>> (dx, x, prev_x, n, 1);
	//__minusMatMat<<<dimGrid_n, dimBlock>>> (dx, x, prev_x, n, 1);
	//Adx = resid - prev_resid;
	__minusMatMat<<<dimGrid_k, dimBlock, 0, st[1]>>> (Adx, resid, prev_resid, k, 1);
	//__minusMatMat<<<dimGrid_k, dimBlock>>> (Adx, resid, prev_resid, k, 1);

	cudaThreadSynchronize();

        //dd  = dx(:)'*dx(:);  
        //dGd = Adx(:)'*Adx(:);
	FLOAT dd, dGd;
	dd = inner(dx, dx, n);
	dGd = inner(Adx, Adx, k);
        alpha = fminf( alphamax, fmaxf(alphamin, dGd/(REALMIN+dd)) );
    }

    /* update iteration count */
    iterThisCycle = (int)fmodf(iterThisCycle+1, bbCycle);

    if (stopCriterion==1)
    {
    // relative change in function value
	FLOAT criterionObjective = fabsf(f-prev_f)/(prev_f);
        keep_going = (criterionObjective > stop_tol ? 1:0);
#ifdef VERBOSE
	mexPrintf("Delta obj. = %e \n", criterionObjective);
#endif 
    } else if (stopCriterion==2) {
    // duality test. Matlab code:
    // scaleFactor = norm(gradq(:),inf);
    //        w = tau*prev_resid(:) / scaleFactor;
    //        criterionDuality = 0.5* (prev_resid(:)'*prev_resid(:)) + ...
    //            tau * phi_function(prev_x) + 0.5*w(:)'*w(:) + y(:)'*w(:);
    //        criterionDuality = criterionDuality / prev_f;
    //        keep_going = (criterionDuality > tolA);

    FLOAT scaleFactor = norm_inf(gradq, n);
    FLOAT criterionDuality =  
      0.5f*(1.f+(tau/scaleFactor)*(tau/scaleFactor))*prev_residTresid +
      tau*prev_norm1x +
      (tau/scaleFactor)*inner(y, prev_resid, k);
    criterionDuality = criterionDuality / prev_f;
    keep_going = (criterionDuality > stop_tol ? 1:0);
#ifdef VERBOSE
	mexPrintf("Duality Gap = %e \n", criterionDuality);
#endif 
    }

    if (iter > maxiter) {
	keep_going = 0;
    }

  } // end of while(keep_going)
#ifdef VERBOSE
    mexPrintf("Number of non-zero components = %d\n",my_nnz(x,n));
#endif

} //end of continuation loop

exec_time = cputime() - exec_time;

    /* Print results */
#ifdef VERBOSE
    mexPrintf("Number of iterations = %d\n",iter);
    mexPrintf("0.5*||A x - y ||_2^2 = %10.3e\n",0.5*inner(resid, resid, k));
    mexPrintf("tau * Penalty = %10.3e\n", tau*norm_1(x,n));
    mexPrintf("Objective function = %10.3e\n",f);
    mexPrintf("Number of non-zero components = %d\n",my_nnz(x,n));
    mexPrintf("Total time %f sec, Time per iteration %f sec\n", exec_time, exec_time/iter);
    mexPrintf("\n");
#endif

    /* Return values to Matlab */
    plhs[0] = mxCreateNumericMatrix(n, 1, mxSINGLE_CLASS, mxREAL);
    x_h = (FLOAT*) mxGetData(plhs[0]);
    CUDA_CALL( cudaMemcpy(x_h, x, sizeof(FLOAT)*n, cudaMemcpyDeviceToHost) );

    /*
     * Cleanup GPU variables
     */
    CUDA_CALL( cudaFree(x) );
    CUDA_CALL( cudaFree(prev_x) );
    CUDA_CALL( cudaFree(dx) );
    CUDA_CALL( cudaFree(Aty) );
    CUDA_CALL( cudaFree(gradq) );

    CUDA_CALL( cudaFree(y) );
    CUDA_CALL( cudaFree(resid) );
    CUDA_CALL( cudaFree(prev_resid) );
    CUDA_CALL( cudaFree(Adx) );

    CUDA_CALL( cudaFree(dct_rows) );

    dct_gpu_destroy();
    cudaMath_destroy();

    free(taus);

    cudaStreamDestroy(st[0]);
    cudaStreamDestroy(st[1]);

turn_time = cputime() - turn_time;

    plhs[1] = mxCreateDoubleScalar((double)iter);
    plhs[2] = mxCreateDoubleScalar(turn_time);
    plhs[3] = mxCreateDoubleScalar(exec_time);

}



