function [x,iter] = SpaRSA_short(y,x_start,A,AT,tau,continuationSteps,stop_tol)

% SpaRSA: short version, preparing for GPU implementation, July 2008.
%
% revised to add continuation , October 2008
% 
% arg min_x = 0.5*|| y - A x ||_2^2 + tau phi(x)
%
% using the SpaRSA algorithm, which is described in "Sparse Reconstruction
% by Separable Approximation" by S. J. Wright, R. D. Nowak, and
% M. A. T. Figueiredo, submitted, December, 2007.
% The algorithm is similar to GPSR (Figueiredo, Nowak, Wright) but does not
% rely on the conversion to QP form of the l1 norm, because it is not
% limited to being used with l1 regularization. Instead it forms a separable
% approximation to the first term of the objective, which has the form
%
%  d'*A'*(A x - y) + 0.5*alpha*d'*d 
%
% where alpha is obtained from a Barzilai-Borwein formula. 
% In a monotone variant, alpha is increased until we see a decrease in 
% the original objective function over this step. 
%
% Copyright (2007): Mario Figueiredo, Robert Nowak, Stephen Wright
%
%  ===== Required inputs =============
%
%  y: 1D vector or 2D array (image) of observations.
%
%  x_start: initial value for x.
%     
%  A: handle to a function that computes
%     products of the form A*v, for some vector v whose size is the same as
%     that of x.
%
%  AT: handle to a function that computes 
%     products of the form A^T*u, for some vector u whose size is the same as
%     that of y
%
%  tau: regularization parameter (scalar)
%
%  continuationSteps: number of steps to take in continuation scheme
%
%  stop_tol: tolerance for stopping criterion
%
% ============ Outputs ==============================
% 
%   x = solution of the main algorithm
%
%   objective = sequence of values of the objective function
%
% ========================================================

maxiter = 1000; % max iterations
bbCycle = 1;   % cycle length for Barzilai-Borwein
enforceMonotone = 0;  % enforce monotonicity?
stopCriterion = 1; % 1=function value test; 2=duality test.
alphamin = 1e-30;
alphamax = 1e30;
verbose = 0;   % set to 1 if you want chatty output
eta = 2.0;     % increase alpha by this factor after unsuccessful step
alphaFactor = 0.8;  % decrease factor between iterations, if BB not used
  
  sigma = 1.0; % controls shape of the continuation curve
  
% Precompute A'*y since it'll be used a lot
Aty = AT(y);
Aty_max=norm(Aty(:),inf);

nContinuation=max(continuationSteps,1);
taus = zeros(nContinuation,1);
% compute roster of tau values for continuation
if (tau>=.8*Aty_max) | (nContinuation==1)
  taus(1) = tau;
else
  taus(1) = .8*Aty_max; 
  logtau0 = log(taus(1)); logtaum = log(tau);
  for i=2:nContinuation
    tauFac = (nContinuation-i)/(nContinuation-1.0);
    tauFac = tauFac^sigma;
    logtau = logtaum + tauFac*(logtau0-logtaum);
    taus(i) = exp(logtau);
  end
end


% Initialization
x = x_start;

% start the clock
start_time = cputime();
iter=0;

for j=1:length(taus)
  tau=taus(j);
  % iteration count for this cycle of BB
  iterThisCycle = 0;

  % initialize alpha
  alpha=1.0;
  
  % Compute the initial residual and gradient
  resid =  A(x) - y;
  gradq = AT(resid);
  
  %f = 0.5*(resid(:)'*resid(:)) + tau * norm(x,1);
  residTresid = resid(:)'*resid(:); norm1x = norm(x(:),1);
  f = 0.5*residTresid + tau * norm1x;
  
  if verbose
    fprintf(1,'\n continuation step %2d, tau=%7.3e, iter=%4d, obj=%10.6e\n', ...
	j, tau, iter, f);
  end
  
  % control variable for the outer loop and iteration counter
  keep_going = 1;
  while keep_going
    iter=iter+1;
    gradq = AT(resid);

    % save current values
    prev_x = x;
    prev_f = f;
    prev_resid = resid;
    prev_residTresid = residTresid;
    prev_norm1x = norm1x;
  
    % computation of step
    cont_inner = 1;
    while cont_inner
      x = soft(prev_x - gradq*(1/alpha),tau/alpha);
      %    dx = x - prev_x;
      resid = A(x)-y;
      %    Adx = resid - prev_resid;
      residTresid = resid(:)'*resid(:); norm1x=norm(x(:),1);
      f = 0.5*residTresid + tau * norm1x;
      % did the function decrease? do we care?
      if (f <= prev_f) | (~enforceMonotone)
	cont_inner=0;
      else
	% not good enough, increase alpha and try again
	alpha = eta*alpha;
      end
    end   % of while cont_inner
  
    % print stuff
    if verbose
      fprintf(1,'t=%4d, obj=%10.6e, alpha=%6.3e ', iter, f, alpha );
    end
    
    % standard BB choice of initial alpha for next step
    if iterThisCycle==0 | enforceMonotone==1
      dx = x - prev_x;
      Adx = resid - prev_resid;
      dd  = dx(:)'*dx(:);  
      dGd = Adx(:)'*Adx(:);
      alpha = min(alphamax,max(alphamin,dGd/(realmin+dd)));
    end
    
    % update iteration counts, store results and times
    iterThisCycle=mod(iterThisCycle+1,bbCycle);
    
    if stopCriterion==1
      criterionObjective = abs(f-prev_f)/(prev_f);
      keep_going =  (criterionObjective > stop_tol);
      if verbose
	fprintf(1,'Delta obj. = %e \n', criterionObjective)
      end
    elseif stopCriterion==2
      %    scaleFactor = norm(gradq(:),inf);
      %    w = tau*prev_resid(:) / scaleFactor;
      %    criterionDuality = 0.5* (prev_resid(:)'*prev_resid(:)) + ...
      %	tau * norm(prev_x,1) + 0.5*w(:)'*w(:) + y(:)'*w(:);
      scaleFactor = norm(gradq(:), inf);
      criterionDuality = 0.5*(1.0+(tau/scaleFactor)*(tau/scaleFactor))*prev_residTresid + ...
	  tau*prev_norm1x + ...
	  (tau/scaleFactor)*(y(:)'*prev_resid(:));
      criterionDuality = criterionDuality / prev_f;
      keep_going = (criterionDuality > stop_tol);
      if verbose
	fprintf(1,' rel dgap = %6.3e\n', criterionDuality)
      end
    end
    
    if iter > maxiter
      keep_going = 0;
    end
    
  end % end of the main loop of the GPBB algorithm (while keep_going)
  
  if verbose
    fprintf(1,'number of nonzero components = %d\n', length(find(x~=0)))
  end
  
end

elapse_time = cputime() - start_time;
  
% Print results
if verbose
  fprintf(1,'Number of iterations = %d\n',iter)
  fprintf(1,'0.5*||A x - y ||_2^2 = %10.3e\n',0.5*resid(:)'*resid(:))
  fprintf(1,'tau * Penalty = %10.3e\n', tau*norm(x,1));
  fprintf(1,'Objective function = %10.3e\n',f);
  fprintf(1,'Number of non-zero components = %d\n',sum(x(:)~=0));
  fprintf(1,'Total time %f sec, Time per iteration %f sec\n', ...
      elapse_time, elapse_time/iter);
  fprintf(1,'\n');
end

