/*
 * function [u,w1,w2,Energy,Dgap,itr,total_time,exec_time] = TV_PDHD(w1,w2,f,lbd,NIT,GapTol)
 *
 * GPU (CUDA) implementation of image denoising algorithm, PDHG.
 *
 * Author: Sangkyun Lee
 * Last Modified: 2008/11/7
 */

#include <time.h>
#include "mex.h"
#include "cudaMath.cu"

#define MATLAB		// debug() work as mexPrintf()
// comment out VERBOSE and SAFE_MODE for speed measurements
//#define VERBOSE
//#define SAFE_MODE

void print_usage()
{
    debug("[u,w1,w2,Energy,Dgap,itr,total_time,exec_time] = TV_PDHD(w1,w2,f,lbd,NIT,GapTol)\n");
}

/*
 * Declarations
 *
 */
inline void energy(FLOAT* result, const FLOAT* DivW, const FLOAT* f, const FLOAT* g, FLOAT lbd, int m, int n, cudaStream_t st=0);
inline void primal_f(FLOAT* result, const FLOAT* u, const FLOAT* f, const FLOAT* ux, const FLOAT* uy, FLOAT lbd, int m, int n, cudaStream_t st=0);
__global__ void __normalize(FLOAT* w1, FLOAT* w2, int m, int n);
__global__ void __updateDual(FLOAT* u, const FLOAT* f, const FLOAT* DivW, FLOAT theta, FLOAT rlbd, int m, int n);

/*
 * Main Entry Function for Matlab
 *
 */
//function [u, w1,w2, Energy, Dgap, itr, total_time, exec_time] = TV_PDHD(w1,w2,f,lbd,NIT,GapTol);
void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[])
{
    double turn_time, exec_time;

    FLOAT *w1_h, *w2_h, *f_h;		// in host
    FLOAT *w1, *w2, *f, *g, *DivW;	// in device
    FLOAT *u, *ux, *uy;
    FLOAT lbd, GapTol;
    FLOAT Primal, Dual, Dgap;
    int NIT;
    int m,n;
    int memsize;

    // output variables
    FLOAT *u_out, *w1_out, *w2_out;

    // opt. loop variables
    int itr;
    FLOAT tau, theta;

    if(nrhs != 6) {
	print_usage();
	return;
    }

    // set device
    //CUDA_CALL( cudaSetDevice(2) );	// 9800

turn_time = cputime();

    cudaMath_init();

cudaStream_t stream[2];
cudaStreamCreate(&stream[0]);
cudaStreamCreate(&stream[1]);

    // input variables
    w1_h = (FLOAT*) mxGetData(prhs[0]);		// these arguments should be single()
    w2_h = (FLOAT*) mxGetData(prhs[1]);
    f_h = (FLOAT*) mxGetData(prhs[2]);
    lbd = (FLOAT) mxGetScalar(prhs[3]);
    NIT = (int) mxGetScalar(prhs[4]);
    GapTol = (FLOAT) mxGetScalar(prhs[5]);

    m = mxGetM(prhs[2]);
    n = mxGetN(prhs[2]);
    memsize = sizeof(FLOAT)*m*n;

    // allocate GPU memory
    CUDA_CALL( cudaMalloc((void**)&w1, memsize) );
    CUDA_CALL( cudaMalloc((void**)&w2, memsize) );
    CUDA_CALL( cudaMalloc((void**)&f, memsize) );
    CUDA_CALL( cudaMalloc((void**)&g, memsize) );
    CUDA_CALL( cudaMalloc((void**)&DivW, memsize) );
    CUDA_CALL( cudaMalloc((void**)&u, memsize) );
    CUDA_CALL( cudaMalloc((void**)&ux, memsize) );
    CUDA_CALL( cudaMalloc((void**)&uy, memsize) );

    // copy input vectors to GPU memory
    CUDA_CALL( cudaMemcpy(f, f_h, memsize, cudaMemcpyHostToDevice) );
    CUDA_CALL( cudaMemcpy(w1, w1_h, memsize, cudaMemcpyHostToDevice) );
    CUDA_CALL( cudaMemcpy(w2, w2_h, memsize, cudaMemcpyHostToDevice) );
    CUDA_CALL( cudaMemcpy(u, f, memsize, cudaMemcpyDeviceToDevice) );

    // bind textures
    CUDA_CALL( cudaBindTexture(0, tex, u, memsize) );
    CUDA_CALL( cudaBindTexture(0, tex1, w1, memsize) );
    CUDA_CALL( cudaBindTexture(0, tex2, w2, memsize) );

    // define block and grid
    dim3 dimBlock(BLK_LEN_2D, BLK_LEN_2D);
    dim3 dimGrid(idivup(m,BLK_LEN_2D), idivup(n,BLK_LEN_2D));

exec_time = cputime();

    __timesMatConst<<<dimGrid, dimBlock>>> (g, f, lbd, m, n);

    divergence2(DivW, w1, w2, m, n);
    energy(&Dual, DivW, f, g, lbd, m, n);
    gradient2(ux, uy, u, m, n);
    primal_f(&Primal, u, f, ux, uy, lbd, m, n);

    Dgap = (Primal-Dual) / (fabs(Primal) + fabs(Dual));

#ifdef VERBOSE
    debug(" Initial: Pri=%8.3e, Dua=%8.3e, Gap=%5.2e\n", Primal, Dual, Dgap);
#endif

for (itr=1; itr<=NIT; ++itr)
{
    // choose tau
    tau = 0.2 + 0.08*itr;
    __update<<<dimGrid, dimBlock, 0, stream[0]>>> (w1, ux, -tau*lbd, m, n);	// w1 = w1 - tau*lbd*ux;
    __update<<<dimGrid, dimBlock, 0, stream[1]>>> (w2, uy, -tau*lbd, m, n);	// w2 = w2 - tau*lbd*uy;
    //cudaThreadSynchronize();

   __normalize<<<dimGrid, dimBlock>>> (w1, w2, m, n);

    divergence2(DivW, w1, w2, n, n);
    energy(&Dual, DivW, f, g, lbd, m, n, stream[0]);

    theta = (0.5 - 5.0/(15.0+itr)) / tau;
    __updateDual<<<dimGrid, dimBlock, 0, stream[1]>>> (u, f, DivW, theta, 1.f/lbd, m, n);
    gradient2(ux, uy, u, m, n, stream[1]);

    primal_f(&Primal, u, f, ux, uy, lbd, m, n, stream[1]);
  
    cudaThreadSynchronize();
  
    Dgap = (Primal-Dual) / (fabs(Primal) + fabs(Dual));
  
#ifdef VERBOSE
    debug(" itr %4d: Pri=%8.3e, Dua=%8.3e, Gap=%5.2e\n", itr, Primal, Dual, Dgap);
#endif

    if ( Dgap < GapTol )
	break;
}

exec_time = cputime() - exec_time;

    // handle return values to Matlab
    plhs[0] = mxCreateNumericMatrix(m, n, mxSINGLE_CLASS, mxREAL);
    plhs[1] = mxCreateNumericMatrix(m, n, mxSINGLE_CLASS, mxREAL);
    plhs[2] = mxCreateNumericMatrix(m, n, mxSINGLE_CLASS, mxREAL);
    u_out = (FLOAT*) mxGetData(plhs[0]);
    w1_out = (FLOAT*) mxGetData(plhs[1]);
    w2_out = (FLOAT*) mxGetData(plhs[2]);

    CUDA_CALL( cudaMemcpy(u_out, u, memsize, cudaMemcpyDeviceToHost) );
    CUDA_CALL( cudaMemcpy(w1_out, w1, memsize, cudaMemcpyDeviceToHost) );
    CUDA_CALL( cudaMemcpy(w2_out, w2, memsize, cudaMemcpyDeviceToHost) );

    plhs[3] = mxCreateDoubleScalar((double) Dual);
    plhs[4] = mxCreateDoubleScalar((double) Dgap);
    plhs[5] = mxCreateDoubleScalar((double) itr);

    // free GPU memory
    CUDA_CALL( cudaFree(w1) );
    CUDA_CALL( cudaFree(w2) );
    CUDA_CALL( cudaFree(f) );
    CUDA_CALL( cudaFree(g) );
    CUDA_CALL( cudaFree(DivW) );
    CUDA_CALL( cudaFree(u) );
    CUDA_CALL( cudaFree(ux) );
    CUDA_CALL( cudaFree(uy) );

cudaStreamDestroy(stream[0]);
cudaStreamDestroy(stream[1]);

    cudaMath_destroy();

turn_time = cputime() - turn_time;

    plhs[6] = mxCreateDoubleScalar((double) turn_time);
    plhs[7] = mxCreateDoubleScalar((double) exec_time);

    return;
}

////////////////////////////////////////////////////////////////
// Objective functions
__global__ void __energy_F(FLOAT* par_sum, const FLOAT* DivW, const FLOAT* f, const FLOAT* g, const FLOAT rlbdsq, INT length)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_idx = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_idx < length)
        cache[ty][tx] = __fsqr_rn(f[lin_idx]) - __fmul_rn(rlbdsq, __fsqr_rn(DivW[lin_idx] - g[lin_idx]));
    __syncthreads();
    if(lin_idx +256< length)
        cache[ty][tx] += __fsqr_rn(f[lin_idx+256]) - __fmul_rn(rlbdsq, __fsqr_rn(DivW[lin_idx+256] - g[lin_idx+256]));
    __syncthreads();
    if(lin_idx +512< length)
        cache[ty][tx] += __fsqr_rn(f[lin_idx+512]) - __fmul_rn(rlbdsq, __fsqr_rn(DivW[lin_idx+512] - g[lin_idx+512]));
    __syncthreads();
    if(lin_idx +768< length)
        cache[ty][tx] += __fsqr_rn(f[lin_idx+768]) - __fmul_rn(rlbdsq, __fsqr_rn(DivW[lin_idx+768] - g[lin_idx+768]));
    __syncthreads();


    if(lin_idx +1024< length)
        cache[ty][tx] += __fsqr_rn(f[lin_idx+1024]) - __fmul_rn(rlbdsq, __fsqr_rn(DivW[lin_idx+1024] - g[lin_idx+1024]));
    __syncthreads();
    if(lin_idx +1280< length)
        cache[ty][tx] += __fsqr_rn(f[lin_idx+1280]) - __fmul_rn(rlbdsq, __fsqr_rn(DivW[lin_idx+1280] - g[lin_idx+1280]));
    __syncthreads();
    if(lin_idx +1536< length)
        cache[ty][tx] += __fsqr_rn(f[lin_idx+1536]) - __fmul_rn(rlbdsq, __fsqr_rn(DivW[lin_idx+1536] - g[lin_idx+1536]));
    __syncthreads();
    if(lin_idx +1792< length)
        cache[ty][tx] += __fsqr_rn(f[lin_idx+1792]) - __fmul_rn(rlbdsq, __fsqr_rn(DivW[lin_idx+1792] - g[lin_idx+1792]));
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] += cache[ty][tx+8];
        cache[ty][tx] += cache[ty][tx+4];
        cache[ty][tx] += cache[ty][tx+2];
        cache[ty][tx] += cache[ty][tx+1];
        __syncthreads();

        if(tx==0 && ty==0) {
	    float sum = 0.f;
	    for(int i=0; i<BLK_LEN_2D;++i)
		sum += cache[i][0];
	    par_sum[blockIdx.x] = sum;
	}
    }
}

inline void energy(FLOAT* result, const FLOAT* DivW, const FLOAT* f, const FLOAT* g, FLOAT lbd, int m, int n, cudaStream_t st)
{
    FLOAT sum_h;

    int block_size = 8*BLK_LEN_2D*BLK_LEN_2D;
    int sum_size = m*n;
    int reduced_size = idivup(sum_size, block_size);

    dim3 dimBlock(BLK_LEN_2D, BLK_LEN_2D);
    dim3 dimGrid(reduced_size, 1);

    // first reduction
    __energy_F <<<dimGrid, dimBlock, 0, st>>> (CM_par_sum, DivW, f, g, 1./(lbd*lbd), m*n);
    sum_size = reduced_size;
    reduced_size = idivup(reduced_size, block_size);
    dimGrid.x = reduced_size;
    dimGrid.y = 1;

    while(sum_size > 1)
    {
	__nill_sum <<<dimGrid, dimBlock, 0, st>>> (CM_par_sum, CM_par_sum, sum_size);

	sum_size = reduced_size;
	reduced_size = idivup(reduced_size, block_size);
	dimGrid.x = reduced_size;
    }

    CUDA_CALL( cudaMemcpy(&sum_h, CM_par_sum, sizeof(FLOAT), cudaMemcpyDeviceToHost) );\

    result[0] = sum_h * lbd / 2.f;
}


__global__ void __primal_F(FLOAT* par_sum, const FLOAT* u, const FLOAT* f, const FLOAT* ux, const FLOAT* uy, const FLOAT hlbd, INT length)
{
    __shared__ FLOAT cache[BLK_LEN_2D][BLK_LEN_2D+1];

    int tx = threadIdx.x;
    int ty = threadIdx.y;
    int lin_idx = __imul(__imul(8*BLK_LEN_2D,BLK_LEN_2D), blockIdx.x) + __imul(BLK_LEN_2D, ty) + tx;

    cache[ty][tx] = 0.f;
    if(lin_idx < length)
        cache[ty][tx] = hypotf(ux[lin_idx],uy[lin_idx]) + __fmul_rn(hlbd, __fsqr_rn(u[lin_idx] - f[lin_idx]));
    __syncthreads();
    if(lin_idx+256 < length)
        cache[ty][tx] += hypotf(ux[lin_idx+256],uy[lin_idx+256]) + __fmul_rn(hlbd, __fsqr_rn(u[lin_idx+256] - f[lin_idx+256]));
    __syncthreads();
    if(lin_idx+512 < length)
        cache[ty][tx] += hypotf(ux[lin_idx+512],uy[lin_idx+512]) + __fmul_rn(hlbd, __fsqr_rn(u[lin_idx+512] - f[lin_idx+512]));
    __syncthreads();
    if(lin_idx+768 < length)
        cache[ty][tx] += hypotf(ux[lin_idx+768],uy[lin_idx+768]) + __fmul_rn(hlbd, __fsqr_rn(u[lin_idx+768] - f[lin_idx+768]));
    __syncthreads();

    if(lin_idx+1024 < length)
        cache[ty][tx] += hypotf(ux[lin_idx+1024],uy[lin_idx+1024]) + __fmul_rn(hlbd, __fsqr_rn(u[lin_idx+1024] - f[lin_idx+1024]));
    __syncthreads();
    if(lin_idx+1280 < length)
        cache[ty][tx] += hypotf(ux[lin_idx+1280],uy[lin_idx+1280]) + __fmul_rn(hlbd, __fsqr_rn(u[lin_idx+1280] - f[lin_idx+1280]));
    __syncthreads();
    if(lin_idx+1536 < length)
        cache[ty][tx] += hypotf(ux[lin_idx+1536],uy[lin_idx+1536]) + __fmul_rn(hlbd, __fsqr_rn(u[lin_idx+1536] - f[lin_idx+1536]));
    __syncthreads();
    if(lin_idx+1792 < length)
        cache[ty][tx] += hypotf(ux[lin_idx+1792],uy[lin_idx+1792]) + __fmul_rn(hlbd, __fsqr_rn(u[lin_idx+1792] - f[lin_idx+1792]));
    __syncthreads();

    if(tx < 8) {
	cache[ty][tx] += cache[ty][tx+8];
        cache[ty][tx] += cache[ty][tx+4];
        cache[ty][tx] += cache[ty][tx+2];
        cache[ty][tx] += cache[ty][tx+1];
        __syncthreads();

        if(tx==0 && ty==0) {
	    float sum = 0.f;
	    for(int i=0; i<BLK_LEN_2D;++i)
		sum += cache[i][0];
	    par_sum[blockIdx.x] = sum;
	}
    }
}
inline void primal_f(FLOAT* result, const FLOAT* u, const FLOAT* f, const FLOAT* ux, const FLOAT* uy, FLOAT lbd, int m, int n, cudaStream_t st)
{
    FLOAT sum_h;

    int block_size = 8*BLK_LEN_2D*BLK_LEN_2D;
    int sum_size = m*n;
    int reduced_size = idivup(sum_size, block_size);

    dim3 dimBlock(BLK_LEN_2D, BLK_LEN_2D);
    dim3 dimGrid(reduced_size, 1);

    // first reduction
    __primal_F <<<dimGrid, dimBlock, 0, st>>> (CM_par_sum, u, f, ux, uy, lbd/2.f, m*n);
    sum_size = reduced_size;
    reduced_size = idivup(reduced_size, block_size);
    dimGrid.x = reduced_size;
    dimGrid.y = 1;

    while(sum_size > 1)
    {
	__nill_sum <<<dimGrid, dimBlock, 0, st>>> (CM_par_sum, CM_par_sum, sum_size);

	sum_size = reduced_size;
	reduced_size = idivup(reduced_size, block_size);
	dimGrid.x = reduced_size;
    }

    CUDA_CALL( cudaMemcpy(&sum_h, CM_par_sum, sizeof(FLOAT), cudaMemcpyDeviceToHost) );\

    result[0] = sum_h;
}

/////////////////////////////////////////////////
// Utility functions
__global__ void __normalize(FLOAT* w1, FLOAT* w2, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;
    int lin_idx = __imul(m,Y) + X;

    FLOAT w1_e = w1[lin_idx];
    FLOAT w2_e = w2[lin_idx];
    FLOAT norm = hypotf(w1_e, w2_e);

    if(norm > 1.0f)
    {
	w1[lin_idx] = __fdividef(w1_e, norm);
	w2[lin_idx] = __fdividef(w2_e, norm);
    }
}

//u= (1.0-theta)*u + theta*(f - (1/lbd)*DivW);   
__global__ void __updateDual(FLOAT* u, const FLOAT* f, const FLOAT* DivW, FLOAT theta, FLOAT rlbd, int m, int n)
{
    int X = X_INDEX;
    int Y = Y_INDEX;

    if(X < m && Y < n)
    {
	int lin_idx = __imul(m,Y) + X;
	float diff = f[lin_idx] - __fmul_rn(rlbd,DivW[lin_idx]);
	u[lin_idx] = __fmul_rn(1.0f-theta, u[lin_idx]) + __fmul_rn(theta, diff);
    }
}



