%Last Updated on 05-21-2008
%Start modifying for deblurring by SKL, 10-15-2008

% This program applies Mingqiang's PDHD algorithm to solve the
% ROF image restoration model

%Dual Formulation of TV model:
% min || div w - \lbd f|| subject to |w| <= 1

% \div : divergence,  \g : gradient
%-------------------------------------------------------------------------
% Input variables
%-------------------------------------------------------------------------
% w1,w2:        Dual variable, initial guess. 
% f:            noisy image
% h:            blur kernel (shifted and padded, real)
% lbd:          Constant fidelity parameter. 
% NIT:          Maximum number of iterations
% GapTol:       Convergence tolerance (stop criterion) for relative duality gap
%--------------------------------------------------------------------------
%-------------------------------------------------------------------------
% Output variables
%-------------------------------------------------------------------------
% u:                Primal variable, numerical solution - restored image 
% w = (w1,w2):      Dual variable, numerical solution
% TimeCost:        CPU time cost
% itr:              number of iterations

function [u, w1,w2, TimeCost, itr] = TV_PDHD(w1,w2,f,h,lbd,NIT,GapTol);

verbose = 0;

t0 = cputime;                %Start CPU clock

h = fft2(h);
ch = conj(h);
chf = ch.*fft2(f);
hh = ch.*h;

n=length(f);                %Assume a square image        

%Compute the primal u
u  = f;
ux = [u(:,2:n)-u(:,1:n-1), zeros(n,1)];
uy = [u(2:n,:)-u(1:n-1,:); zeros(1,n)];

TimeCost = 0;

for itr=1:NIT
  
  % choose tau and update dual w
  tau = 10. + 40.*itr;
  w1 = w1 + tau*ux; w2 = w2 + tau*uy;
  wnorm= max(1, sqrt(w1.^2+w2.^2));
  w1 = w1./wnorm;
  w2 = w2./wnorm;

  DivW = ([w1(:,1),w1(:,2:n)-w1(:,1:n-1)] + [w2(1,:);w2(2:n,:)-w2(1:n-1,:)]); 
  
  % choose theta and update primal u
  theta = (1-.2/itr)/tau;
  new_u = real( ifft2( (fft2(u - theta*DivW) + theta*lbd*chf ) ./ (1 + theta*lbd*hh) ) );

  u = u - new_u;
  tol = norm(u(:), inf) / 256;
  u = new_u;

  if verbose
    fprintf(' itr %4d: tol=%5.2e\n', itr, tol);
  end

  if( tol < GapTol )
    break;
  end

  ux = [u(:,2:n)-u(:,1:n-1), zeros(n,1)];
  uy = [u(2:n,:)-u(1:n-1,:);zeros(1,n)];

end

%Energy = Dual;
TimeCost = cputime-t0;
