/*
 * 1-D and 2-D Discrete Consine Transform using CUFFT
 *
 * All input signal dimension lengths has to be even numbers.
 *
 * Author: Sangkyun Lee
 * Date: 2008/11/7
 *
 */

#include "cuda.h"
#include "cudaMath.cu"

#define PIO2 CUDART_PIO2
#define BLK_DCT_1D 32
#define BLK_DCT_2D 16

int dct_gpu_init_1D = 0;
int dct_gpu_init_2D = 0;

#define DCT_SAFE_MODE

#ifdef DCT_SAFE_MODE
    #define CHECK_INIT_1D()	{ if(!dct_gpu_init_1D){ debug("Error: dct_gpu is not initialized for 1-D transforms!\n"); exit(-1);}}
    #define CHECK_INIT_2D()	{ if(!dct_gpu_init_2D){ debug("Error: dct_gpu is not initialized for 2-D transforms!\n"); exit(-1);}}
#else
    #define CHECK_INIT_1D()	{}
    #define CHECK_INIT_2D()	{}
#endif

// Global variables
Complex *DCT_y, *DCT_fft_y;
FLOAT *DCT_whole_x, *DCT_whole_out, *tmp;
cufftHandle DCT_plan_row, DCT_plan_col;

// Initialization
void dct_gpu_init(int M, int N)
{
    int MN = M*N;

    CUDA_CALL( cudaMalloc((void**)&DCT_y, sizeof(Complex)*MN) );
    CUDA_CALL( cudaMalloc((void**)&DCT_fft_y, sizeof(Complex)*MN) );
    CUDA_CALL( cudaMalloc((void**)&DCT_whole_x, sizeof(FLOAT)*MN) );
    CUDA_CALL( cudaMalloc((void**)&DCT_whole_out, sizeof(FLOAT)*MN) );

    // fft plan for column-wise transform
    CUFFT_CALL( cufftPlan1d(&DCT_plan_col, M, CUFFT_C2C, N) );
    dct_gpu_init_1D = 1;

    if(N > 1) {
	// fft plan for row-wise transform
	CUFFT_CALL( cufftPlan1d(&DCT_plan_row, N, CUFFT_C2C, M) );
	CUDA_CALL( cudaMalloc((void**)&tmp, sizeof(FLOAT)*MN) );
	dct_gpu_init_2D = 1;
    }
}

// Deallocation
void dct_gpu_destroy()
{
    CUDA_CALL( cudaFree(DCT_y) );
    CUDA_CALL( cudaFree(DCT_fft_y) );
    CUDA_CALL( cudaFree(DCT_whole_x) );
    CUDA_CALL( cudaFree(DCT_whole_out) );
    CUFFT_CALL( cufftDestroy(DCT_plan_col) );
    dct_gpu_init_1D = 0;

    if(dct_gpu_init_2D) {
        CUFFT_CALL( cufftDestroy(DCT_plan_row) );
        CUDA_CALL( cudaFree(tmp) );
	dct_gpu_init_2D = 0;
    }
}

/*
 * Utility functions to manipulate signals
 *
 */
// 1-D
__global__ void __split(Complex* y, const FLOAT* x, INT N)
{
    int m = X_INDEX;
    Complex c;

    if(m < N/2)
    {
	c.x = x[2*m];
	c.y = 0.f;
	y[m] = c;
    }
    else if(m < N)
    {
	m -= N/2;
	c.x = x[2*m+1];
	c.y = 0.f;
	y[N-m-1] = c;
    }
}
// 2-D
__global__ void __split2(Complex* y, const FLOAT* x, INT M, INT N)
{
    int m = X_INDEX;
    int n = Y_INDEX;

    if(m < M && n < N) {
	int base = M*n;
	Complex c;

        if(m < M/2)
        {
	    c.x = x[2*m + base];
	    c.y = 0.f;
	    y[m + base] = c;
	}
	else
	{
	    m -= M/2;
	    c.x = x[2*m+1 + base];
	    c.y = 0.f;
	    y[M-m-1 + base] = c;
	}
    }
}
// 1-D
__global__ void __gather(FLOAT* x, const Complex* y, INT N)
{
    int m = X_INDEX;

    if(m < N/2)
	x[2*m] = y[m].x;
    else if(m < N)
    {
	m -= N/2;
	x[2*m+1] = y[N-m-1].x;
    }
}
// 2-D
__global__ void __gather2(FLOAT* x, const Complex* y, INT M, INT N)
{
    int m = X_INDEX;
    int n = Y_INDEX;

    if(n < N) {
	int base = M*n;

        if(m < M/2)
	    x[2*m + base] = y[m + base].x;
	else if(m < M)
	{
	    m -= M/2;
	    x[2*m+1 + base] = y[M-m-1 + base].x;
	}
    }
}
// 1-D
__global__ void __fft2dct(FLOAT* out, Complex* in, INT N)
{
    int n = X_INDEX;
    FLOAT w;
    FLOAT f = n*PIO2/N;

    /*
    if(n < N/2+1)
	c = in[n];
    else if(n < N)
    {
	c = in[N-n];
	c.y *= -1.f;
    }
    */

    if(n < N) {
	Complex c = in[n];

	if(n==0)
	    w = sqrtf(1.f/N);
	else
	    w = sqrtf(2.f/N);

	out[n] = w*(c.x*__cosf(f) + c.y*__sinf(f));
    }
}
// 2-D
__global__ void __fft2dct2(FLOAT* out, Complex* in, INT M, INT N)
{
    int m = X_INDEX;
    int n = Y_INDEX;

    if(m < M && n < N) {
	FLOAT w;
	FLOAT f = m*PIO2/M;
	INT base = M*n;
	Complex c = in[m + base];

	if(m==0)
	    w = sqrtf(1.f/M);
	else
	    w = sqrtf(2.f/M);

	out[m + base] = w*(c.x*__cosf(f) + c.y*__sinf(f));
    }
}
// 1-D
__global__ void __dct2fft(Complex* out, const FLOAT* in, INT N)
{
    int n = X_INDEX;

    if(n < N) {
        FLOAT f = n*PIO2/N;
	FLOAT d = in[n];
        FLOAT w;
        Complex c;

        if(n==0)
	    w = sqrtf(1.f/N);
	else
	    w = sqrtf(2.f/N);

	c.x = w*d*__cosf(f);
	c.y = w*d*__sinf(f);

	out[n] = c;
    }
}
// 2-D
__global__ void __dct2fft2(Complex* out, const FLOAT* in, INT M, INT N)
{
    int m = X_INDEX;
    int n = Y_INDEX;

    if(m < M && n < N) {
        FLOAT f = m*PIO2/M;
	INT base = M*n;
	FLOAT d = in[m + base];
        FLOAT w;
        Complex c;

        if(m==0)
	    w = sqrtf(1.f/M);
	else
	    w = sqrtf(2.f/M);

	c.x = w*d*__cosf(f);
	c.y = w*d*__sinf(f);

	out[m + base] = c;
    }
}

/*
 * Main routines.
 *
 */

//
// FORWARD TRANSFORMS
//

/*
 * Perform 1-D DCT on the signal x of length N
 */
void dct_gpu(FLOAT* out, const FLOAT* x, INT N)
{
    CHECK_INIT_1D();
//    Complex* y;
//    Complex* fft_y;

//    CUDA_CALL( cudaMalloc((void**)&y, sizeof(Complex)*N) );
//    CUDA_CALL( cudaMalloc((void**)&fft_y, sizeof(Complex)*N) );

    dim3 dimBlock(BLK_DCT_1D, 1);
    dim3 dimGrid((int)ceilf(1.f*N/BLK_DCT_1D), 1);

    __split<<<dimGrid, dimBlock>>> (DCT_y, x, N);

    // do fft on y
//    cufftHandle plan;
//    CUFFT_CALL( cufftPlan1d(&plan, N, CUFFT_C2C, 1) );
    CUFFT_CALL( cufftExecC2C(DCT_plan_col, (cufftComplex*)DCT_y, (cufftComplex*)DCT_fft_y, CUFFT_FORWARD) );	// CUFFT_FORWARD

    __fft2dct<<<dimGrid, dimBlock>>> (out, DCT_fft_y, N);

    // cleanup
//    CUFFT_CALL( cufftDestroy(plan) );
//    CUDA_CALL( cudaFree(y) );
//    CUDA_CALL( cudaFree(fft_y) );
///    cudaThreadSynchronize();
}

/*
// A very slow CUDA version of 2-D DCT
void dct2_gpu(FLOAT* out, const FLOAT* x, INT M, INT N)
{
    CHECK_INIT_2D();

    register int i;

    // colume-wise dct
    for(i=0; i<N; ++i) {
	//dct_gpu(out + M*i, x + M*i, M);
	dim3 dct_block(DCT_BLOCK, 1);
	dim3 dct_grid((int)ceilf(1.f*M/DCT_BLOCK), 1);
	__split<<<dct_grid, dct_block>>> (DCT_y, x + M*i, M);
	CUFFT_CALL( cufftExecC2C(DCT_plan_col, (cufftComplex*)DCT_y, (cufftComplex*)DCT_fft_y, CUFFT_FORWARD) );
	__fft2dct<<<dct_grid, dct_block>>> (out + M*i, DCT_fft_y, M);
    }

    // transpose
    dim3 grid((int)ceilf(1.f*M/BLOCK_DIM), int(ceilf(1.f*N/BLOCK_DIM)));
    dim3 threads(BLOCK_DIM, BLOCK_DIM);
    __transpose<<<grid, threads>>> (out, out, M, N);
    
    // row-wise dct
    for(i=0; i<M; ++i) {
	//dct_gpu(out + N*i, x + N*i, N);
	dim3 dct_block(DCT_BLOCK, 1);
	dim3 dct_grid((int)ceilf(1.f*N/DCT_BLOCK), 1);
	__split<<<dct_grid, dct_block>>> (DCT_y, out + N*i, N);
	CUFFT_CALL( cufftExecC2C(DCT_plan_row, (cufftComplex*)DCT_y, (cufftComplex*)DCT_fft_y, CUFFT_FORWARD) );
	__fft2dct<<<dct_grid, dct_block>>> (out + N*i, DCT_fft_y, N);
    }

    // transpose back
    swapi(grid.x, grid.y);
    __transpose<<<grid, threads>>> (out, out, N, M);
}
*/

/*
 * Perform 2-D DCT on 2-D signal x of dimension M by N.
 *
 */
void dct2_gpu(FLOAT* out, const FLOAT* x, INT M, INT N)
{
    CHECK_INIT_2D();

    dim3 dct_block(BLK_DCT_2D, BLK_DCT_2D);
    dim3 dct_grid((int)ceilf(1.f*M/BLK_DCT_2D), (int)ceilf(1.f*N/BLK_DCT_2D));

    // colume-wise dct
    __split2<<<dct_grid, dct_block>>> (DCT_y, x, M, N);
    CUFFT_CALL( cufftExecC2C(DCT_plan_col, (cufftComplex*)DCT_y, (cufftComplex*)DCT_fft_y, CUFFT_FORWARD) );
    __fft2dct2<<<dct_grid, dct_block>>> (out, DCT_fft_y, M, N);

    // transpose
    dim3 grid((int)ceilf(1.f*M/BLK_LEN_2D), (int)ceilf(1.f*N/BLK_LEN_2D));
    dim3 threads(BLK_LEN_2D, BLK_LEN_2D);
    //__transpose<<<grid, threads>>> (out, out, M, N);
    __transpose<<<grid, threads>>> (tmp, out, M, N);

    // row-wise dct
    swapi(dct_grid.x, dct_grid.y);
    __split2<<<dct_grid, dct_block>>> (DCT_y, tmp, N, M);
    CUFFT_CALL( cufftExecC2C(DCT_plan_row, (cufftComplex*)DCT_y, (cufftComplex*)DCT_fft_y, CUFFT_FORWARD) );
    __fft2dct2<<<dct_grid, dct_block>>> (tmp, DCT_fft_y, N, M);

    // transpose back
    swapi(grid.x, grid.y);
    __transpose<<<grid, threads>>> (out, tmp, N, M);
}

/*
 * Functions for dealing with sampled rows of DCT(x)
 */
__global__ void __select_rows(FLOAT* out, const FLOAT* in, INT* rows, INT k)
{
    int X = X_INDEX;

    if(X < k)
	out[X] = in[rows[X]-1];			// -1 is because the indices are from Matlab
}
__global__ void __fill_rows(FLOAT* out, const FLOAT* in, INT *rows, INT k)
{
    int X = X_INDEX;

    if(X < k)
	out[rows[X]-1] = in[X];
}

/*
 * Perform 1-D DCT and return only the selected rows of DCT(x).
 */
// output has k elements, selected by *rows rows
void dct_gpu(FLOAT* out, const FLOAT* x, INT N, INT k, INT* rows)
{
    CHECK_INIT_1D();

//    FLOAT *whole_out;

//    CUDA_CALL( cudaMalloc((void**)&whole_out, sizeof(FLOAT)*N) );

    dct_gpu(DCT_whole_out, x, N);

    dim3 dimBlock(BLK_DCT_1D, 1);
    dim3 dimGrid((int)fmaxf(1.f, ceilf(1.f*k/BLK_DCT_1D)), 1);
    __select_rows<<<dimGrid, dimBlock>>>(out, DCT_whole_out, rows, k);

//    CUDA_CALL( cudaFree(whole_out) );
///    cudaThreadSynchronize();
}

/*
 * Perform 2-D DCT and return only the selected rows of DCT(x).
 */
void dct2_gpu(FLOAT* out, const FLOAT* x, INT m, INT n, INT k, INT* rows)
{
    CHECK_INIT_2D();

    dct2_gpu(DCT_whole_out, x, m, n);

    dim3 dimBlock(BLK_DCT_1D, 1);
    dim3 dimGrid((int)fmaxf(1.f, ceilf(1.f*k/BLK_DCT_1D)), 1);
    __select_rows<<<dimGrid, dimBlock>>>(out, DCT_whole_out, rows, k);

///    cudaThreadSynchronize();
}




//
// INVERSE TRANSFORMS
//

/*
 * Perform 1-D inverse DCT transform
 */
void idct_gpu(FLOAT* out, const FLOAT* x, INT N)
{
    CHECK_INIT_1D();

//    Complex* y;
//    Complex* fft_y;

//    CUDA_CALL( cudaMalloc((void**)&y, sizeof(Complex)*N) );
//    CUDA_CALL( cudaMalloc((void**)&fft_y, sizeof(Complex)*N) );
    dim3 dimBlock(BLK_DCT_1D, 1);
    dim3 dimGrid((int)fmaxf(1.f, (int)ceilf(1.f*N/BLK_DCT_1D)), 1);

    __dct2fft<<<dimGrid, dimBlock>>> (DCT_fft_y, x, N);

    // do ifft on fft_y
//    cufftHandle plan;
//    CUFFT_CALL( cufftPlan1d(&plan, N, CUFFT_C2C, 1) );
    CUFFT_CALL( cufftExecC2C(DCT_plan_col, (cufftComplex*)DCT_fft_y, (cufftComplex*)DCT_y, CUFFT_INVERSE) ); // CUFFT_INVERSE

    __gather<<<dimGrid, dimBlock>>> (out, DCT_y, N);

    // cleanup
//    CUFFT_CALL( cufftDestroy(plan) );
//    CUDA_CALL( cudaFree(y) );
//    CUDA_CALL( cudaFree(fft_y) );

///    cudaThreadSynchronize();
}

/*
 * Perform 2-D inverse DCT transform
 */
void idct2_gpu(FLOAT* out, const FLOAT* x, INT M, INT N)
{
    CHECK_INIT_2D();

    dim3 dct_block(BLK_DCT_2D, BLK_DCT_2D);
    dim3 dct_grid((int)ceilf(1.f*M/BLK_DCT_2D), (int)ceilf(1.f*N/BLK_DCT_2D));

    // colume-wise dct
    __dct2fft2<<<dct_grid, dct_block>>> (DCT_fft_y, x, M, N);
    CUFFT_CALL( cufftExecC2C(DCT_plan_col, (cufftComplex*)DCT_fft_y, (cufftComplex*)DCT_y, CUFFT_INVERSE) );
    __gather2<<<dct_grid, dct_block>>> (out, DCT_y, M, N);

    // transpose
    dim3 grid((int)ceilf(1.f*M/BLK_LEN_2D), (int)ceilf(1.f*N/BLK_LEN_2D));
    dim3 threads(BLK_LEN_2D, BLK_LEN_2D);
    __transpose<<<grid, threads>>> (tmp, out, M, N);

    // row-wise dct
    swapi(dct_grid.x, dct_grid.y);
    __dct2fft2<<<dct_grid, dct_block>>> (DCT_fft_y, tmp, N, M);
    CUFFT_CALL( cufftExecC2C(DCT_plan_row, (cufftComplex*)DCT_fft_y, (cufftComplex*)DCT_y, CUFFT_INVERSE) );
    __gather2<<<dct_grid, dct_block>>> (tmp, DCT_y, N, M);

    // transpose back
    swapi(grid.x, grid.y);
    __transpose<<<grid, threads>>> (out, tmp, N, M);

///    cudaThreadSynchronize();
}

/*
 * Perform 1-D inverse DCT transform, for a set of sampled input rows.
 */
// input x has k elements, selected by *rows rows
void idct_gpu(FLOAT* out, const FLOAT* x, INT N, INT k, INT* rows)
{
    CHECK_INIT_1D();

//    FLOAT *whole_x;

//    CUDA_CALL( cudaMalloc((void**)&whole_x, sizeof(FLOAT)*N) );
    CUDA_CALL( cudaMemset(DCT_whole_x, 0, sizeof(FLOAT)*N) );

    dim3 dimBlock(BLK_DCT_1D, 1);
    dim3 dimGrid((int)fmaxf(1.f, ceilf(1.f*k/BLK_DCT_1D)), 1);
    __fill_rows<<<dimGrid, dimBlock>>>(DCT_whole_x, x, rows, k);

    idct_gpu(out, DCT_whole_x, N);
//    CUDA_CALL( cudaFree(whole_x) );
}

/*
 * Perform 2-D inverse DCT transform, for a set of sampled input rows.
 */
void idct2_gpu(FLOAT* out, const FLOAT* x, INT m, INT n, INT k, INT* rows)
{
    CHECK_INIT_2D();

    CUDA_CALL( cudaMemset(DCT_whole_x, 0, sizeof(FLOAT)*m*n) );

    dim3 dimBlock(BLK_DCT_1D, 1);
    dim3 dimGrid((int)fmaxf(1.f, ceilf(1.f*k/BLK_DCT_1D)), 1);
    __fill_rows<<<dimGrid, dimBlock>>>(DCT_whole_x, x, rows, k);

    idct2_gpu(out, DCT_whole_x, m, n);
}

