/*
 * function gpu_warmup()
 *
 * This code is for selecting and warming-up a GPU.
 *
 * The very first CUDA call, the very first memory allocation,
 * and the very first computation call takes comparably long
 * time.
 *
 * Call this routine with the device ID number you want to use
 * for the current process (e.g. Matlab), and to do some dummy
 * tasks in order to avoid these from affecting the runtime of
 * algorithms.
 *
 * This function lists all CUDA-enabled devices in the system
 * if it is called without any arguments.
 *
 * Author: Sangkyun Lee
 * Last Modified: 2008/11/7
 */

#include "mex.h"
#include "../cudaMath/cuda.h"

#define MATLAB
#define CHORE_SIZE (1<<12)

void print_usage()
{
    debug("\ngpu_warmup(device, verbose)\n");
    debug("  If no argument is set, list CUDA-capable devices on system\n");
    debug("  Otherwise, the arguments mean:\n");
    debug("   - device: device id to be used for this thread\n");
    debug("   - verbose: if 1, it prints out how much time were taken for one-shot chores (setDevice,etc)\n");
}

void list_device_caps(int dev, cudaDeviceProp& dp)
{
    debug("\nDevice %d: \"%s\"\n", dev, dp.name);
    debug("  Revision                       %d.%d\n", dp.major, dp.minor);
    debug("  Global memory                  %u bytes\n", dp.totalGlobalMem);
    debug("  Constant memory                %u bytes\n", dp.totalConstMem);
    debug("  Shared memory per block        %u bytes\n", dp.sharedMemPerBlock);
    debug("  # registers per block          %d\n", dp.regsPerBlock);
    debug("  # multiprocessors              %d\n", dp.multiProcessorCount);
    debug("  # cores                        %d\n", 8*dp.multiProcessorCount);
    debug("  Warp size                      %d\n", dp.warpSize);
    debug("  Max # of threads per block     %d\n", dp.maxThreadsPerBlock);
    debug("  Max dim of a block             %d x %d x %d\n", dp.maxThreadsDim[0], dp.maxThreadsDim[1], dp.maxThreadsDim[2]);
    debug("  Max dim of a grid              %d x %d x %d\n", dp.maxGridSize[0], dp.maxGridSize[1], dp.maxGridSize[2]);
    debug("  Max memory pitch               %u bytes\n", dp.memPitch);
    debug("  Texture alignment              %u bytes\n", dp.textureAlignment);
    debug("  Clock rate                     %.2f GHz\n", dp.clockRate * 1e-6f);
    debug("  Concurrent copy and execution: %s\n", dp.deviceOverlap? "Yes":"No");
}

/*
 * Some dummy task (add two matrices)
 */
__global__ void __plus_chore(FLOAT* out, const FLOAT* w1, const FLOAT* w2, int m, int n)
{
    int X = __mul24(blockIdx.x,blockDim.x) + threadIdx.x;
    int Y = __mul24(blockIdx.y,blockDim.y) + threadIdx.y;
    int lin_idx;

    if(X < m && Y < n)
    {
	lin_idx = __mul24(m,Y) + X;
	out[lin_idx] = w1[lin_idx] + w2[lin_idx];
    }
}


void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, const mxArray *prhs[])
{
    float* tmp;
    int num_devices;
    int dev;

    int requested_device;
    int verbose = 0;

    double firstcuda_time, setdev_time, alloc_time, firstcall_time, secondcall_time;

    firstcuda_time = cputime();
    CUDA_CALL( cudaGetDeviceCount(&num_devices) );
    firstcuda_time = cputime() - firstcuda_time;

    if(nrhs < 1) {
	if(num_devices == 0) {
	    debug("There is no device supporting CUDA\n");
	    return;
	} else {
	    debug("There are %d devices supporting CUDA\n", num_devices);
	}

	// list devices and their capabilities
	for(dev = 0; dev < num_devices; ++dev) {
	    cudaDeviceProp dp;
	    CUDA_CALL( cudaGetDeviceProperties(&dp, dev) );
	    list_device_caps(dev, dp);
	}
	return;
    }

    requested_device = (int) mxGetScalar(prhs[0]);
    if(nrhs > 1)
	verbose = (int) mxGetScalar(prhs[1]);

    if(requested_device >= num_devices) {
	debug("Error: there are %d CUDA devices: set device id [%d,%d]\n", num_devices, 0, max(0,num_devices-1));
	return;
    }

    // set device
    setdev_time = cputime();
    CUDA_CALL( cudaSetDevice(requested_device) );
    setdev_time = cputime() - setdev_time;

// Do one-shot chores
    // the very first memory allocation
    alloc_time = cputime();
    CUDA_CALL( cudaMalloc((void**)&tmp, sizeof(float)*CHORE_SIZE) );
    alloc_time = cputime() - alloc_time;

    // call some math routines on tmp memory to heat up GPU
    dim3 dimBlock(128,1);
    dim3 dimGrid(CHORE_SIZE/128,1);

    // the very first run on GPU (GPU seems to be cooled down to a lower clock rate)
    firstcall_time = cputime();
    __plus_chore<<<dimBlock, dimGrid>>> (tmp, tmp, tmp, CHORE_SIZE, 1);
    cudaThreadSynchronize();
    firstcall_time = cputime() - firstcall_time;

    secondcall_time = cputime();
    __plus_chore<<<dimBlock, dimGrid>>> (tmp, tmp, tmp, CHORE_SIZE, 1);
    cudaThreadSynchronize();
    secondcall_time = cputime() - secondcall_time;

    CUDA_CALL( cudaFree(tmp) );

    if(verbose) {
	debug("\nTime taken for GPU warmup\n");
	debug("  The very first cuda call:     %.4fs\n", firstcuda_time);
	debug("  The set device call:          %.4fs\n", setdev_time);
	debug("  The very first allocation:    %.4fs\n", alloc_time);
	debug("  Speed-up by warmup            %.1fx\n", firstcall_time/secondcall_time);
    }
}


